/**
 * File for general "Load Balancing" logic.
 */


// User Import Headers.
#ifndef user_headers
    #define user_headers
    #include "load_balance_schemes.h"
    #include "load_balance_general.h"
    #include "structs.h"
    #include "terminal_commands.h"
#endif


/**
 * Entrypoint for logic to run "Asynchronous Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void arr_run(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("\n");
        printf("Starting ARR load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        arr_worker(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        arr_main(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("Finished ARR load scheme.\n");
        printf("\n");
    }
}


/**
 * Main processor logic for "Asynchronous Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void arr_main(thread_struct* thread_args_ptr) {
    int index;
    int main_loop_bool = 3;
    int msg_received = 0;
    int* recv_array;

    log("Starting ARR main.");

    // Initialize console display.
    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        main_display_status(thread_args_ptr, recv_array[0], recv_array[1], 1);
        free(recv_array);
    }

    // Continually grab messages from all worker processors until done.
    // For each loop, it checks all worker processes for a message about work being done.
    // If no messages are found for two consecutive loops in a row, then assumes all work is complete and terminates.
    index = 1;
    while (main_loop_bool > 0) {

        // Check status of worker process.
        if (main_check_worker_status(thread_args_ptr, index)) {
            msg_received = 1;
        }

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            // Check if messages were recieved this run.
            if (msg_received == 0) {
                // No messages recieved.
                main_loop_bool -= 1;

                // Send termination request if loop bool hit 0.
                // Otherwise loop a few more times to give requests time to propagate.
                if (main_loop_bool <= 0) {
                    main_terminate_workers(thread_args_ptr);
                }

            } else {
                // One or more messages recieved this loop. Reset main loop value.
                main_loop_bool = 3;
            }

            // Sleep number of seconds provided by user terminal.
            // Meant to simulate "work", as it forces each worker process to block for this amount of time each load,
            // before continuing on. In a real program, the worker would be doing something useful in this time,
            // instead of just blocking.
            sleep(thread_args_ptr->seconds_per_load);
            msg_received = 0;
        }

    }

    log("Finished ARR main.");
}


/**
 * Worker processor logic for "Asynchronous Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void arr_worker(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int arr_counter = thread_args_ptr->thread_num;
    int main_loop_bool = 1;
    int msg_status_flag;
    int request_flag = -1;
    int work_value = 0;
    int send_status = 1;

    log("Starting ARR worker.");

    // Initialize process "1" to have full loads. All others start with none.
    if (process_rank == 1) {
        thread_args_ptr->remaining_loads = thread_args_ptr->total_loads;
    } else {
        thread_args_ptr->remaining_loads = 0;
    }

    // Initialize output by immediately sending status to main process.
    worker_send_status(process_rank, thread_args_ptr->remaining_loads);

    // Main loop bool. Continues until main process says all work is done.
    while (main_loop_bool == 1) {

        // Loop all worker processes until load count message of "0" has been sent.
        while (thread_args_ptr->remaining_loads > 0) {

            // Send message for current load.
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);

            // Check for message requests from other processes.
            msg_status_flag = 0;
            MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // At least one message pending. Handle before proceeding.
                worker_handle_request(thread_args_ptr);
            }

            // Decriment load number.
            thread_args_ptr->remaining_loads -= 1;
            send_status = 1;
        }

        // If we made it this far, then worker no longer has work.

        // Send notification message to main, to show value of 0.
        if (send_status == 1) {
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);
            send_status = 0;
        }

        // Check for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }


        if (request_flag == -1) {
            // Update personal ARR counter.
            arr_counter = (arr_counter + 1) % thread_args_ptr->total_processors;
            while (arr_counter == thread_args_ptr->thread_num || arr_counter == 0) {
                arr_counter = (arr_counter + 1) % thread_args_ptr->total_processors;
            }

            // Beg fellow worker for work.
            worker_send_request(thread_args_ptr, arr_counter);
            request_flag = arr_counter;
        } else {
            msg_status_flag = 0;
            MPI_Iprobe(request_flag, tag_work_response, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                MPI_Recv(&work_value, 1, MPI_INT, request_flag, tag_work_response, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Update personal load value if non-reject message provided.
                if (work_value > 0) {
                    thread_args_ptr->remaining_loads = work_value;
                }

                // Reset flag to allow begging for work again in the future.
                request_flag = -1;
            }
        }

        // Check if program as a whole is done running.
        main_loop_bool = worker_check_status();

        // Check again for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        // Check status of overall work.
        if (main_loop_bool == 1) {
            // Sleep to simulate "doing work".
            sleep(1);
        }
    }

    log("Finished ARR worker.");
}


/**
 * Entrypoint for logic to run "Global Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void grr_run(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("\n");
        printf("Starting GRR load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        grr_worker(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        grr_main(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("Finished GRR load scheme.\n");
        printf("\n");
    }
}


/**
 * Main processor logic for "Global Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void grr_main(thread_struct* thread_args_ptr) {
    int index;
    int grr_index;
    int main_loop_bool = 3;
    int msg_received = 0;
    int* recv_array;
    int msg_status_flag;
    int grr_counter = 0;
    int temp = 0;
    MPI_Request request_var;

    log("Starting GRR main.");

    // Initialize console display.
    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        main_display_status(thread_args_ptr, recv_array[0], recv_array[1], 1);
        free(recv_array);
    }

    // Continually grab messages from all worker processors until done.
    // For each loop, it checks all worker processes for a message about work being done.
    // If no messages are found for two consecutive loops in a row, then assumes all work is complete and terminates.
    index = 1;
    while (main_loop_bool > 0) {

        // Check status of worker process.
        if (main_check_worker_status(thread_args_ptr, index)) {
            msg_received = 1;
        }

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            // Check if messages were recieved this run.
            if (msg_received == 0) {
                // No messages recieved.
                main_loop_bool -= 1;

                // Send termination request if loop bool hit 0.
                // Otherwise loop a few more times to give requests time to propagate.
                if (main_loop_bool <= 0) {
                    main_terminate_workers(thread_args_ptr);
                }

            } else {
                // One or more messages recieved this loop. Reset main loop value.
                main_loop_bool = 3;
            }

            // Handle any pending requests for new counter value.
            for (grr_index = 1; grr_index < thread_args_ptr->total_processors; grr_index++) {

                // Check if specific process is requesting counter.
                msg_status_flag = 0;
                MPI_Iprobe(grr_index, tag_grr_query, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
                if (msg_status_flag == 1) {
                    // Officially receive message. We don't care about contents though.
                    MPI_Recv(&temp, 1, MPI_INT, grr_index, tag_grr_query, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                    grr_counter = (grr_counter + 1) % thread_args_ptr->total_processors;
                    while (grr_counter <= 0 || grr_counter == grr_index) {
                        grr_counter = (grr_counter + 1) % thread_args_ptr->total_processors;
                    }

                    // Send counter to worker.
                    MPI_Isend(&grr_counter, 1, MPI_INT, grr_index, tag_grr_query, MPI_COMM_WORLD, &request_var);
                    char* log_msg = calloc(256, sizeof(char));
                    sprintf(log_msg, "Sent counter of %i to processor %i.", grr_counter, grr_index);
                    log(log_msg); free(log_msg);
                }

            }

            // Sleep number of seconds provided by user terminal.
            // Meant to simulate "work", as it forces each worker process to block for this amount of time each load,
            // before continuing on. In a real program, the worker would be doing something useful in this time,
            // instead of just blocking.
            sleep(thread_args_ptr->seconds_per_load);
            msg_received = 0;
        }

    }

    log("Finished GRR main.");
}


/**
 * Worker processor logic for "Global Round Robin" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void grr_worker(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int grr_counter = thread_args_ptr->thread_num;
    int main_loop_bool = 1;
    int msg_status_flag;
    int request_flag = -1;
    int work_value = 0;
    int send_status = 1;
    int waiting_on_grr_resp = 0;
    MPI_Request request_var;

    log("Starting GRR worker.");

    // Initialize process "1" to have full loads. All others start with none.
    if (process_rank == 1) {
        thread_args_ptr->remaining_loads = thread_args_ptr->total_loads;
    } else {
        thread_args_ptr->remaining_loads = 0;
    }

    // Initialize output by immediately sending status to main process.
    worker_send_status(process_rank, thread_args_ptr->remaining_loads);

    // Main loop bool. Continues until main process says all work is done.
    while (main_loop_bool == 1) {

        // Loop all worker processes until load count message of "0" has been sent.
        while (thread_args_ptr->remaining_loads > 0) {

            // Send message for current load.
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);

            // Check for message requests from other processes.
            msg_status_flag = 0;
            MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // At least one message pending. Handle before proceeding.
                worker_handle_request(thread_args_ptr);
            }

            // Decriment load number.
            thread_args_ptr->remaining_loads -= 1;
            send_status = 1;
        }

        // If we made it this far, then worker no longer has work.

        // Send notification message to main, to show value of 0.
        if (send_status == 1) {
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);
            send_status = 0;
        }

        // Check for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }


        if (request_flag == -1) {
            // Update personal GRR counter or send request to thread if counter received.

            if (waiting_on_grr_resp == 0) {
                // Updates counter by sending request to process 0.
                // (Holding a value in process 0 is the closest I know to achieving a "shared memory" value through
                // distributed systems).
                MPI_Isend(&waiting_on_grr_resp, 1, MPI_INT, 0, tag_grr_query, MPI_COMM_WORLD, &request_var);
                waiting_on_grr_resp = 1;

                log("Sent counter request to processor 0.");

            } else {
                msg_status_flag = 0;
                MPI_Iprobe(0, tag_grr_query, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
                if (msg_status_flag == 1) {
                    MPI_Recv(&grr_counter, 1, MPI_INT, 0, tag_grr_query, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                    char* log_msg = calloc(256, sizeof(char));
                    sprintf(log_msg, "Got counter of %i from processor 0.", grr_counter);
                    log(log_msg); free(log_msg);

                    // Beg fellow worker for work.
                    worker_send_request(thread_args_ptr, grr_counter);
                    request_flag = grr_counter;
                    waiting_on_grr_resp = 0;
                }
            }


        } else {
            msg_status_flag = 0;
            MPI_Iprobe(request_flag, tag_work_response, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                MPI_Recv(&work_value, 1, MPI_INT, request_flag, tag_work_response, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Update personal load value if non-reject message provided.
                if (work_value > 0) {
                    thread_args_ptr->remaining_loads = work_value;
                }

                // Reset flag to allow begging for work again in the future.
                request_flag = -1;
            }
        }

        // Check if program as a whole is done running.
        main_loop_bool = worker_check_status();

        // Check again for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        // Check status of overall work.
        if (main_loop_bool == 1) {
            // Sleep to simulate "doing work".
            sleep(1);
        }
    }

    log("Finished GRR worker.");
}


/**
 * Entrypoint for logic to run "Random Poling" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void rp_run(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("\n");
        printf("Starting RP load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        rp_worker(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        rp_main(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("Finished RP load scheme.\n");
        printf("\n");
    }
}


/**
 * Main processor logic for "Random Poling" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void rp_main(thread_struct* thread_args_ptr) {
    int index;
    int main_loop_bool = 3;
    int msg_received = 0;
    int* recv_array;

    log("Starting RP main.");

    // Initialize console display.
    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        main_display_status(thread_args_ptr, recv_array[0], recv_array[1], 1);
        free(recv_array);
    }

    // Continually grab messages from all worker processors until done.
    // For each loop, it checks all worker processes for a message about work being done.
    // If no messages are found for two consecutive loops in a row, then assumes all work is complete and terminates.
    index = 1;
    while (main_loop_bool > 0) {

        // Check status of worker process.
        if (main_check_worker_status(thread_args_ptr, index)) {
            msg_received = 1;
        }

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            // Check if messages were recieved this run.
            if (msg_received == 0) {
                // No messages recieved.
                main_loop_bool -= 1;

                // Send termination request if loop bool hit 0.
                // Otherwise loop a few more times to give requests time to propagate.
                if (main_loop_bool <= 0) {
                    main_terminate_workers(thread_args_ptr);
                }

            } else {
                // One or more messages recieved this loop. Reset main loop value.
                main_loop_bool = 3;
            }

            // Sleep number of seconds provided by user terminal.
            // Meant to simulate "work", as it forces each worker process to block for this amount of time each load,
            // before continuing on. In a real program, the worker would be doing something useful in this time,
            // instead of just blocking.
            sleep(thread_args_ptr->seconds_per_load);
            msg_received = 0;
        }

    }

    log("Finished RP main.");
}


/**
 * Worker processor logic for "Random Poling" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void rp_worker(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int rp_counter;
    int main_loop_bool = 1;
    int msg_status_flag;
    int request_flag = -1;
    int work_value = 0;
    int send_status = 1;

    log("Starting RP worker.");

    // Initialize random number generator.
    srand(time(NULL));

    // Initialize process "1" to have full loads. All others start with none.
    if (process_rank == 1) {
        thread_args_ptr->remaining_loads = thread_args_ptr->total_loads;
    } else {
        thread_args_ptr->remaining_loads = 0;
    }

    // Initialize output by immediately sending status to main process.
    worker_send_status(process_rank, thread_args_ptr->remaining_loads);

    // Main loop bool. Continues until main process says all work is done.
    while (main_loop_bool == 1) {

        // Loop all worker processes until load count message of "0" has been sent.
        while (thread_args_ptr->remaining_loads > 0) {

            // Send message for current load.
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);

            // Check for message requests from other processes.
            msg_status_flag = 0;
            MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // At least one message pending. Handle before proceeding.
                worker_handle_request(thread_args_ptr);
            }

            // Decriment load number.
            thread_args_ptr->remaining_loads -= 1;
            send_status = 1;
        }

        // If we made it this far, then worker no longer has work.

        // Send notification message to main, to show value of 0.
        if (send_status == 1) {
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);
            send_status = 0;
        }

        // Check for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }


        if (request_flag == -1) {
            // Update personal RP counter.
            rp_counter = -1;
            while (rp_counter < 1 || rp_counter >= thread_args_ptr->total_processors) {
                rp_counter = rand() % thread_args_ptr->total_processors;
            }

            // Beg fellow worker for work.
            worker_send_request(thread_args_ptr, rp_counter);
            request_flag = rp_counter;
        } else {
            msg_status_flag = 0;
            MPI_Iprobe(request_flag, tag_work_response, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                MPI_Recv(&work_value, 1, MPI_INT, request_flag, tag_work_response, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Update personal load value if non-reject message provided.
                if (work_value > 0) {
                    thread_args_ptr->remaining_loads = work_value;
                }

                // Reset flag to allow begging for work again in the future.
                request_flag = -1;
            }
        }

        // Check if program as a whole is done running.
        main_loop_bool = worker_check_status();

        // Check again for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        // Check status of overall work.
        if (main_loop_bool == 1) {
            // Sleep to simulate "doing work".
            sleep(1);
        }
    }

    log("Finished RP worker.");
}



/**
 * Entrypoint for logic to run "Nearest Neighbor" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void nn_run(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;

    if (process_rank == 0) {
        printf("\n");
        printf("Starting NN load scheme.\n");
        printf("\n");
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    // Handle based on process number.
    if (process_rank > 0) {
        // Child process. Handles all the work.
        nn_worker(thread_args_ptr);
    } else {
        // Main process. Handles communication to terminal.
        nn_main(thread_args_ptr);
    }

    // Wait for all processes to synchronize.
    // Also sleep for a second to allow time for console output to process.
    MPI_Barrier(MPI_COMM_WORLD);
    sleep(1);

    if (process_rank == 0) {
        printf("\n");
        printf("Finished NN load scheme.\n");
        printf("\n");
    }
}


/**
 * Main processor logic for "Nearest Neighbor" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void nn_main(thread_struct* thread_args_ptr) {
    int index;
    int main_loop_bool = 3;
    int msg_received = 0;
    int* recv_array;

    log("Starting NN main.");

    // Initialize console display.
    for (index = 1; index < thread_args_ptr->total_processors; index++) {
        recv_array = calloc(3, sizeof(int));
        MPI_Recv(recv_array, 2, MPI_INT, index, tag_status_update, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        main_display_status(thread_args_ptr, recv_array[0], recv_array[1], 1);
        free(recv_array);
    }

    // Continually grab messages from all worker processors until done.
    // For each loop, it checks all worker processes for a message about work being done.
    // If no messages are found for two consecutive loops in a row, then assumes all work is complete and terminates.
    index = 1;
    while (main_loop_bool > 0) {

        // Check status of worker process.
        if (main_check_worker_status(thread_args_ptr, index)) {
            msg_received = 1;
        }

        // Increment for next thread.
        index = (index + 1) % thread_args_ptr->total_processors;

        // Check if past last processor.
        if (index == 0) {
            index += 1;

            // Check if messages were recieved this run.
            if (msg_received == 0) {
                // No messages recieved.
                main_loop_bool -= 1;

                // Send termination request if loop bool hit 0.
                // Otherwise loop a few more times to give requests time to propagate.
                if (main_loop_bool <= 0) {
                    main_terminate_workers(thread_args_ptr);
                }

            } else {
                // One or more messages recieved this loop. Reset main loop value.
                main_loop_bool = 3;
            }

            // Sleep number of seconds provided by user terminal.
            // Meant to simulate "work", as it forces each worker process to block for this amount of time each load,
            // before continuing on. In a real program, the worker would be doing something useful in this time,
            // instead of just blocking.
            sleep(thread_args_ptr->seconds_per_load);
            msg_received = 0;
        }

    }

    log("Finished NN main.");
}


/**
 * Worker processor logic for "Nearest Neighbor" load scheme.
 *
 * :param thread_args_ptr: Struct for processor values.
 */
void nn_worker(thread_struct* thread_args_ptr) {
    int process_rank = thread_args_ptr->thread_num;
    int nn_counter = process_rank;
    int main_loop_bool = 1;
    int msg_status_flag;
    int request_flag = -1;
    int work_value = 0;
    int send_status = 1;
    int ten_percent;
    int nn_range;
    int nn_range_low = process_rank;
    int nn_range_high = process_rank;
    int calculate_nn;
    int index;

    log("Starting NN worker.");

    // Determine "neighbor" range, based on thread number.
    // Range is [ process_id +- 10%_of_total_processes ]. If hits 0 or max, then loops around.
    ten_percent = thread_args_ptr->total_processors / 10;
    if (ten_percent < 1) {
        ten_percent = 1;
    }
    nn_range = ten_percent * 2;

    for (index = 0; index < nn_range; index++) {
        nn_range_low -= 1;
        if (nn_range_low == 0) {
            nn_range_low = thread_args_ptr->total_processors - 1;
        }

        nn_range_high = (nn_range_high + 1) % thread_args_ptr->total_processors;
        if (nn_range_high == 0) {
            nn_range_high += 1;
        }
    }

    char* log_msg = calloc(256, sizeof(char));
    sprintf(log_msg, "nn_range: %i", nn_range);
    log(log_msg); free(log_msg);
    log_msg = calloc(256, sizeof(char));
    sprintf(log_msg, "nn_range_low: %i", nn_range_low);
    log(log_msg); free(log_msg);
    log_msg = calloc(256, sizeof(char));
    sprintf(log_msg, "nn_range_high: %i", nn_range_high);
    log(log_msg); free(log_msg);

    // Initialize process "1" to have full loads. All others start with none.
    if (process_rank == 1) {
        thread_args_ptr->remaining_loads = thread_args_ptr->total_loads;
    } else {
        thread_args_ptr->remaining_loads = 0;
    }

    // Initialize output by immediately sending status to main process.
    worker_send_status(process_rank, thread_args_ptr->remaining_loads);

    // Main loop bool. Continues until main process says all work is done.
    while (main_loop_bool == 1) {

        // Loop all worker processes until load count message of "0" has been sent.
        while (thread_args_ptr->remaining_loads > 0) {

            // Send message for current load.
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);

            // Check for message requests from other processes.
            msg_status_flag = 0;
            MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                // At least one message pending. Handle before proceeding.
                worker_handle_request(thread_args_ptr);
            }

            // Decriment load number.
            thread_args_ptr->remaining_loads -= 1;
            send_status = 1;
        }

        // If we made it this far, then worker no longer has work.

        // Send notification message to main, to show value of 0.
        if (send_status == 1) {
            worker_send_status(process_rank, thread_args_ptr->remaining_loads);
            send_status = 0;
        }

        // Check for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        if (request_flag == -1) {
            // Update personal NN counter.
            calculate_nn = 1;
            while (calculate_nn == 1) {
                nn_counter = (nn_counter + 1) % thread_args_ptr->total_processors;
                calculate_nn = 0;

                // Check for basic validation.
                if (nn_counter == 0 || nn_counter == thread_args_ptr->thread_num) {
                    calculate_nn = 1;
                }

                // Check for range validation.
                if (nn_range_low < nn_range_high) {
                    // Standard value set range. Handle normally.
                    if (nn_counter < nn_range_low || nn_counter > nn_range_high) {
                        calculate_nn = 1;
                    }
                } else {
                    // Non-standard value set range. There's an overflow.
                    if (nn_range_high < nn_counter && nn_counter < nn_range_low) {
                        calculate_nn = 1;
                    }
                }
            }

            // Beg fellow worker for work.
            worker_send_request(thread_args_ptr, nn_counter);
            request_flag = nn_counter;
        } else {
            msg_status_flag = 0;
            MPI_Iprobe(request_flag, tag_work_response, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
            if (msg_status_flag == 1) {
                MPI_Recv(&work_value, 1, MPI_INT, request_flag, tag_work_response, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

                // Update personal load value if non-reject message provided.
                if (work_value > 0) {
                    thread_args_ptr->remaining_loads = work_value;
                }

                // Reset flag to allow begging for work again in the future.
                request_flag = -1;
            }
        }

        // Check if program as a whole is done running.
        main_loop_bool = worker_check_status();

        // Check again for pending work requests.
        msg_status_flag = 0;
        MPI_Iprobe(MPI_ANY_SOURCE, tag_work_request, MPI_COMM_WORLD, &msg_status_flag, MPI_STATUS_IGNORE);
        if (msg_status_flag == 1) {
            // At least one message pending. Handle before proceeding.
            worker_handle_request(thread_args_ptr);
        }

        // Check status of overall work.
        if (main_loop_bool == 1) {
            // Sleep to simulate "doing work".
            sleep(1);
        }
    }

    log("Finished NN worker.");
}
